/*
 * Ambient - the ultimate desktop
 * ------------------------------
 * Copyright 2001-2005 by David Gerber <zapek@morphos.net>
 * Copyright 2006 Ambient Open Source Team
 * All Rights Reserved
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * $Id: findresultlistclass.c,v 1.8 2013/10/29 22:34:22 geit Exp $
 */

#include "ambient.h"

/* public */
#include <dos/dosextens.h>
#include <proto/dos.h>

/* private */
#include "mui_func.h"
#include "command.h"
#include "rexx.h"
#include "contextmenu.h"
#include "mimetype.h"
#include "actiondispatcherclass.h"
#include "action.h"
#include "typescanner.h"
#include "file_func.h"
#include "methodstack.h"
#include "threads.h"
#include "mimeuri.h"
#include "ambient_cat.h"

#define BUF_TITLE_SIZE 0x30

struct Data {
	ULONG sort_col;
	LONG sort_direction;
	APTR mimetype;
	APTR cmenu;
	TEXT buf_title[BUF_TITLE_SIZE];  /* used for variable name + escaping   */
};

struct item {
	STRPTR filepart;
	TEXT text[ 0 ];
};



static void doset(APTR obj, struct Data *data, struct TagItem *tags)
{
	FORTAG(tags)
	{
		case MUIA_List_TitleClick:
		{
			if(data->sort_col == tag->ti_Data)
				data->sort_direction *= -1;
			else
				data->sort_col = tag->ti_Data;

			DoMethod(obj, MUIM_List_Sort);
			break;
		}

		case MA_Icon_MimeType:
		{
			/* ok, cheating here. will fix laters. */

			data->mimetype = (APTR)tag->ti_Data;
			break;
		}

		case MUIA_List_DoubleClick:
		{
			struct item *item;
			DoMethod(obj, MUIM_List_GetEntry, MUIV_List_GetEntry_Active, &item);

			if (item)
			{
				APTR dispatcher = (APTR)DoMethod(app, MM_Application_CreateActionDispatcher);

				if (dispatcher)
				{
					TEXT path[PATH_SIZE];

					stccpy(path, item->text, sizeof(path));
					AddPart(path, item->filepart, sizeof(path));

					SetAttrs(dispatcher,
						MA_ActionDispatcher_Event, ACTION_EVENT_DOUBLECLICK,
						MA_ActionDispatcher_SrcID, getv(_win( obj ), MA_Window_ID),
						MA_ActionDispatcher_RefWin, _win(obj),
						TAG_DONE
					);

					DoMethod(dispatcher, MM_ActionDispatcher_AddURI, path, TRUE);
					DoMethod(dispatcher, MM_ActionDispatcher_Execute);
				}
			}

			break;
		}
	}
	NEXTTAG
}


DEFNEW
{
	obj = DoSuperNew(cl, obj,
		MUIA_List_Format, "C=0 MIW=-1 BAR,C=1 MIW=-1",
		MUIA_List_Title, TRUE,
		InputListFrame,
		MUIA_CycleChain, TRUE,
		MUIA_ContextMenu, 1,
		TAG_MORE, INITTAGS
	);

	if (obj)
	{
		GETDATA;

		memset(data, 0, sizeof(struct Data));

		data->sort_direction = 1;
	}

	return (ULONG)obj;
}

DEFMMETHOD(List_Construct)
{
	STRPTR path = msg->entry;
	struct item *item = AllocVecPooled(msg->pool, sizeof( struct item ) + strlen( path ) + 2);

	if (item)
	{
		strcpy(item->text, path);
		item->filepart = FilePart(item->text);

		if (item->filepart)
		{
			memmove(item->filepart + 1, item->filepart, strlen( item->filepart ) + 1);
			item->text[ item->filepart - item->text ] = 0;
			item->filepart++;
		}
		else
		{
			item->filepart = item->text;
		}
	}

	return (ULONG)item;
}

DEFMMETHOD(List_Destruct)
{
	if (msg->entry != NULL)
		FreeVecPooled(msg->pool, msg->entry);

	return 0;
}

static LONG compareEntries( struct Data * data UNUSED, struct item *e1, struct item *e2, LONG col )
{
	LONG rc = 0;

	if (col == 0)
	{
		/* column 0 - name */

		rc = stricmp(e1->filepart, e2->filepart);
	}
	else if (col == 1)
	{
		/* column 1 - path */

		rc = stricmp(e1->text, e2->text);

		if (rc == 0)
			rc = stricmp(e1->filepart, e2->filepart);
	}

	return rc;
}

DEFMMETHOD(List_Compare)
{
	GETDATA;

	struct item *e1 = (struct item *)msg->entry1;
	struct item *e2 = (struct item *)msg->entry2;

	/* get clicked col and reverse state*/
	LONG col = data->sort_col;
	LONG rev = data->sort_direction;

	LONG result;

	/* compare for each column */

	result = compareEntries(data, e1 , e2, col);
	result *= rev;

	return result;
}

DEFMMETHOD(List_Display)
{
	GETDATA;

	struct item *item = (struct item*)msg->entry;

	if (!item)
	{
		snprintf( data->buf_title, BUF_TITLE_SIZE, "%s \033I[6:%s]", GSI( MSG_FINDRESULTLISTCLASS_NAME + data->sort_col ), (data->sort_direction > 0) ? "38" : "39" );

		msg->array[ 0 ] = data->sort_col ? GSI( MSG_FINDRESULTLISTCLASS_NAME ) : data->buf_title;
		msg->array[ 1 ] = data->sort_col ? data->buf_title : GSI( MSG_FINDRESULTLISTCLASS_CONTAINER );
	}
	else
	{
		msg->array[ 0 ] = item->filepart;
		msg->array[ 1 ] = item->text;

		if ((ULONG)msg->array[ -1 ] % 2)
			msg->array[ -9 ] = (STRPTR)10;
	}

	return 0;
}

DEFSET
{
	GETDATA;

	doset(obj, data, INITTAGS);
	return (DOSUPER);
}

DEFMMETHOD(ContextMenuBuild)
{
	GETDATA;

	struct MUI_List_TestPos_Result res;

	DoMethod(obj, MUIM_List_TestPos, msg->mx, msg->my, &res);

	if (data->cmenu)
	{
		MUI_DisposeObject(data->cmenu);
		data->cmenu = NULL;
	}

	if(res.entry == -1)
		return NULL;

	{
		struct item *item;

		DoMethod(obj, MUIM_List_GetEntry, res.entry, (ULONG *)&item);

		set(obj, MUIA_List_Active, res.entry);

		if(item)
		{
			if ((data->cmenu = contextmenu_build(CM_LIST, 0)))
			{
				TEXT path[PATH_SIZE];
				APTR mimetype = NULL;

				/*
				 * Add type dependant actions to context menu.
				 */

				stccpy(path, item->text, sizeof(path));
				if (res.column == 0)
					AddPart(path, item->filepart, sizeof(path));

				if (mimetype == NULL)
				{
					/*
					 * Needs to be found. This is a bit problematic as it should be done on
					 * a thread but for now we do this workaround.
					 */

					LONG cnt = 0;
					
					do_action(obj, TA_MimeType_Scan,
									TT_MimeType_Scan_Path, path,
									TT_MimeType_Scan_MimetypePtr, &data->mimetype,
								TAG_DONE);

					/*
					 * We give it max 1s to find the type.
					 */

					while( cnt < 25 && data->mimetype == NULL )
					{
						Delay(2);
						cnt++;
						methodstack_check(FALSE);
					}

					mimetype = data->mimetype;
					data->mimetype = NULL;
				}

				if (mimetype)
				{
					TEXT buf[ 1024 ];
					STRPTR mode = "List";

					{
						UBYTE encpath[mimeuri_encodepath(NULL, 0, path)];

						mimeuri_encodepath(encpath, sizeof(encpath), path);

						/* XXX: if it's a directory, read icon window coords and pass them, or use some listview preferences coords ? */
						snprintf(buf, sizeof(buf), "file:///%s?view=%s&mode=all",
							encpath,
							mode );
					}

					contextmenu_add_mime(data->cmenu, buf, mimetype);
				}

				return ((ULONG)data->cmenu);
			}
		}
	}

	return (NULL);
}

DEFMMETHOD(ContextMenuChoice)
{
	struct command_menu *cm;

	if ((cm = (struct command_menu *)getv(msg->item, MA_Menuitem_Command)))
	{
		if (cm->name && *cm->name)
		{
			execute_command(obj, cm->type, cm->args, NULL); /* XXX: I think.. well no.. we need the object */
		}

		if (cm->actionnode)
		{
			/*
			 * This is action we selected. Needs to be parsed.
			 */

			APTR dispatcher = (APTR)DoMethod( app, MM_Application_CreateActionDispatcher );

			if ( dispatcher )
			{
				SetAttrs( dispatcher,
					MA_ActionDispatcher_SrcID, getv(_win(obj), MA_Window_ID),
					MA_ActionDispatcher_RefWin, _win(obj),
					MA_ActionDispatcher_Action, cm->actionnode,
					TAG_DONE
				);

				DoMethod(dispatcher, MM_ActionDispatcher_AddURI, cm->args, TRUE);
				DoMethod(dispatcher, MM_ActionDispatcher_Execute);
			}
		}
	}
	return (0);
}


BEGINMTABLE
DECNEW
DECSET
DECMMETHOD(List_Compare)
DECMMETHOD(List_Display)
DECMMETHOD(List_Construct)
DECMMETHOD(List_Destruct)
DECMMETHOD(ContextMenuBuild)
DECMMETHOD(ContextMenuChoice)
ENDMTABLE

DECSUBCLASS_NC(MUIC_List, findresultlistclass)
